#ifndef __BORDER_CONDITION_HH__
#define __BORDER_CONDITION_HH__

#include "image.hh"


void set_border_condition_dirichlet(env_t &e, image<float>& ima, const int thread_id)
{
  const int begin_row = e.NR_threads[thread_id];
  const int end_row   = e.NR_threads[thread_id+1];

  const int NC		= ima.ncols();
  const int NB_THREADS = e.NB_THREADS;

  // up border
  if (thread_id == 0)
    for (int j=0; j<NC; ++j)
      ima(-1,j) = 0;

  // left border
  for (int i=begin_row; i<end_row; ++i)
    {
      ima(i,-1) = 0;
    }

  // right border
  for (int i=begin_row; i<end_row; ++i)
    {
      ima(i,NC) = 0;
    }

  // down border
  if (thread_id == (NB_THREADS-1))
    for (int j=0; j<NC; ++j)
      {
	ima(end_row,j) = 0;
      }
}




// void set_border_condition_dirichlet(image<float>& ima, const unsigned int thread_id)
// {
//   const int NR = ima.nrows();
//   const int NC = ima.ncols();

//   // up border
//   for (int j=0; j<NC; ++j)
//     {
//       ima(-1,j) = 0;
//     }

//   // left border
//   for (int i=0; i<NR; ++i)
//     {
//       ima(i,-1) = 0;
//     }

//   // right border
//   for (int i=0; i<NR; ++i)
//     {
//       ima(i,NC) = 0;
//     }

//   // down border
//   for (int j=0; j<NC; ++j)
//     {
//       ima(NR,j) = 0;
//     }
// }


void set_border_condition_copy(image<float>& ima)
{
  const int NR = ima.nrows();
  const int NC = ima.ncols();

  // up border
  for (int j=0; j<NC; ++j)
    {
      ima(-1,j) = ima(0,j);
    }

  // left border
  for (int i=0; i<NR; ++i)
    {
      ima(i,-1) = ima(i,0);
    }

  // right border
  for (int i=0; i<NR; ++i)
    {
      ima(i,NC) = ima(i,NC-1);
    }

  // down border
  for (int j=0; j<NC; ++j)
    {
      ima(NR,j) = ima(NR-1,j);
    }

}



void set_border_condition_laplacian_zero(env_t &e, image<float>& ima, const int thread_id)
{
  const int begin_row = e.NR_threads[thread_id];
  const int end_row   = e.NR_threads[thread_id+1];

  const int NC		= ima.ncols();
  const int NB_THREADS  = e.NB_THREADS;

  // up border
  if (thread_id == 0)
    for (int j=0; j<NC; ++j)
      {
	float steep = ima(1,j) - ima(0,j);
	ima(-1,j) = ima(0,j) - steep; // ima(-1,j) = 2*ima(0,j) - ima(1,j)
      }

  // left border
  for (int i=begin_row; i<end_row; ++i)
    {
      float steep = ima(i,1) - ima(i,0);
      ima(i,-1) = ima(i,0) - steep;
    }

  // right border
  for (int i=begin_row; i<end_row; ++i)
    {
      float steep = ima(i,NC-2) - ima(i,NC-1);
      ima(i,NC) = ima(i,NC-1) - steep;
    }

  // down border
  if (thread_id == (NB_THREADS-1))
    for (int j=0; j<NC; ++j)
      {
	float steep = ima(end_row-2,j) - ima(end_row-1,j);
	ima(end_row,j) = ima(end_row-1,j) - steep;
      }
}



void set_border_condition_landscape_water_constant(env_t &e,
						   image<float>& water,
						   image<float>& landscape,
						   const float cst,
						   const int thread_id)
{
  const int begin_row = e.NR_threads[thread_id];
  const int end_row   = e.NR_threads[thread_id+1];

  const int NC		= water.ncols();
  const int NB_THREADS  = e.NB_THREADS;

  // up border
  if (thread_id == 0)
    for (int j=0; j<NC; ++j)
      {
	if (landscape(-1,j) > cst)
	  water(-1,j) = 0;
	else
	  water(-1,j) =  cst - landscape(-1,j);
      }

  // left border
  for (int i=begin_row; i<end_row; ++i)
    {
      if (landscape(i,-1) > cst)
	water(i,-1) = 0;
	else
	  water(i,-1) =  cst - landscape(i,-1);
    }

  // right border
  for (int i=begin_row; i<end_row; ++i)
    {
      if (landscape(i,NC) > cst)
	water(i,NC) = 0;
	else
	  water(i,NC) =  cst - landscape(i,NC);
    }

  // down border
  if (thread_id == (NB_THREADS-1))
    for (int j=0; j<NC; ++j)
      {
      if (landscape(end_row,j) > cst)
	water(end_row,j) = 0;
      else
	  water(end_row,j) =  cst - landscape(end_row,j);
      }


}

void set_border_condition_ocean(env_t &e,
				image<float>& water,
				image<float>& landscape, const float cst, const int thread_id)
{
  set_border_condition_laplacian_zero(e, landscape, thread_id);
  set_border_condition_landscape_water_constant(e, water, landscape, cst, thread_id);
}



#endif
