#ifndef __LANDSCAPE_EVOLUTION_HH__
#define __LANDSCAPE_EVOLUTION_HH__

#include "image.hh"
#include "env.hh"
#include "border_condition.cc"

#include <cmath>



template<int N>
float my_pow(const float a)
{
  return powf(a, float(N));
}

template<>
float my_pow<1>(const float a)
{
  return a;
}

template<>
float my_pow<2>(const float a)
{
  return a*a;
}



float perform_one_step_landscape_evolution(env_t& e, const int thread_id)
{
  image<float> & height_landscape		 = *(e.height_landscape);
  const image<float> & level_water		 = *(e.level_water);
  const image<float> & water_times_concentration = *(e.water_times_concentration);
  image<float> & variation_landscape		 = *(e.variation_landscape);

  const float DT = e.DT;

  const float EXP_M = e.EROSION_EXP_M;
  const float EROSION_E = e.EROSION_E;
  const float EROSION_S     = e.EROSION_S;


//   const int NB_W = 4;
//   const int win_i[NB_W] = {0,  0, -1, 1};
//   const int win_j[NB_W] = {-1, 1,  0, 0};

  const int begin_row = e.NR_threads[thread_id];
  const int end_row   = e.NR_threads[thread_id+1];
  const int NC		= level_water.ncols();

  float ret = 0;

  for (int i=begin_row; i<end_row; ++i)
    {
      for (int j=0; j<NC; ++j)
	{
	  float amount_to_erode = 0;
	  float h_ij = height_landscape(i,j) + level_water(i,j);
	  { // neighbor (i, j-1)
	    float h_neigh = height_landscape(i,j-1) + level_water(i,j-1);
	    float val_to_exp = fmax(h_ij - h_neigh,0);
	    if (val_to_exp > 0)
	      amount_to_erode += powf(val_to_exp, 2*EXP_M);
	  }
	  { // neighbor (i, j+1)
	    float h_neigh = height_landscape(i,j+1) + level_water(i,j+1);
	    float val_to_exp = fmax(h_ij - h_neigh,0);
	    if (val_to_exp > 0)
	      amount_to_erode += powf(val_to_exp, 2*EXP_M);
	  }
	  { // neighbor (i-1, j)
	    float h_neigh = height_landscape(i-1,j) + level_water(i-1,j);
	    float val_to_exp = fmax(h_ij - h_neigh,0);
	    if (val_to_exp > 0)
	      amount_to_erode += powf(val_to_exp, 2*EXP_M);
	  }
	  { // neighbor (i+1, j)
	    float h_neigh = height_landscape(i+1,j) + level_water(i+1,j);
	    float val_to_exp = fmax(h_ij - h_neigh,0);
	    if (val_to_exp > 0)
	      amount_to_erode += powf(val_to_exp, 2*EXP_M);
	  }

  	  amount_to_erode = -EROSION_E * powf(level_water(i,j), EXP_M) * amount_to_erode ;

	  // rho =
	  float rho = 0.;
	  // EPSILON_WATER = small amount of water
	  const float EPSILON_WATER=1e-8;
	  if ((level_water(i,j) < EPSILON_WATER) || (water_times_concentration(i,j)<0) )
	    rho = 0.;
	  else
	    rho = water_times_concentration(i,j)/level_water(i,j);
	  float sedimentation = fminf( water_times_concentration(i,j), DT*EROSION_S*rho );


	  variation_landscape(i,j) = DT*amount_to_erode + sedimentation;
	} // end j
    } // end i


#pragma omp barrier
  for (int i=begin_row; i<end_row; ++i)
    for (int j=0; j<NC; ++j)
      {
	ret                   += variation_landscape(i,j);
	height_landscape(i,j) += variation_landscape(i,j);
	// we divide by DT because we need z^(n+1)-z^n for the sedimentation evolution
	variation_landscape(i,j) = variation_landscape(i,j) / DT;
      }
#pragma omp barrier

  return ret;
}



float perform_landscape_evolution(env_t &e, const unsigned int thread_id)
{
  // set border condition for water
//   set_border_condition_laplacian_zero(level_water);
//   // set border condition for landscape
//   set_border_condition_laplacian_zero(landscape);

  // set border condition for water
  // set border condition for landscape

//   set_border_condition_dirichlet(e, *(e.level_water), thread_id);

//    set_border_condition_laplacian_zero(e, *(e.level_water), thread_id);
//    set_border_condition_laplacian_zero(e, *(e.height_landscape), thread_id);
//    set_border_condition_dirichlet(e, *(e.height_landscape), thread_id);

  set_border_condition_ocean(e, *(e.level_water), *(e.height_landscape),
			     e.CST_OCEAN_LEVEL, thread_id);


  // perform one step of landscape evolution
   float ret = perform_one_step_landscape_evolution(e, thread_id);
   return ret;
}


#endif
