#ifndef __LANDSCAPE_EVOLUTION_HH__
#define __LANDSCAPE_EVOLUTION_HH__

#include "image.hh"
#include "env.hh"
#include "border_condition.cc"

#include <cmath>

float compute_laplacian(const float cur,
			const float n1, const float n2, const float n3, const float n4)
{
  float output = n1 + n2 + n3 + n4 - 4*cur;
  return output;
}


template<int N>
float my_pow(const float a)
{
  return powf(a, float(N));
}

template<>
float my_pow<1>(const float a)
{
  return a;
}

template<>
float my_pow<2>(const float a)
{
  return a*a;
}



void perform_one_step_landscape_evolution(env_t& e, const int thread_id)
{
  image<float> & height_landscape		 = *(e.height_landscape);
  const image<float> & level_water		 = *(e.level_water);
  const image<float> & water_times_concentration = *(e.water_times_concentration);
  image<float> & variation_landscape		 = *(e.variation_landscape);

  const float DT = e.DT;

  const float EXP_M = e.EROSION_EXP_M;
  const float EROSION_E = e.EROSION_E;
  const float EROSION_CREEP = e.EROSION_CREEP;
  const float EROSION_S     = e.EROSION_S;


  const int NB_W = 4;
  const int win_i[NB_W] = {0,  0, -1, 1};
  const int win_j[NB_W] = {-1, 1,  0, 0};

  const int begin_row = e.NR_threads[thread_id];
  const int end_row   = e.NR_threads[thread_id+1];
  const int NC		= level_water.ncols();

  for (int i=begin_row; i<end_row; ++i)
    {
      for (int j=0; j<NC; ++j)
	{
	  float amount_to_erode = 0;
	  float h_ij = height_landscape(i,j) + level_water(i,j);

	  for (int k=0; k<NB_W; ++k)
	    {
	      int n_i = i+win_i[k];
	      int n_j = j+win_j[k];
	      float h_neigh = height_landscape(n_i,n_j) + level_water(n_i,n_j);
  	      amount_to_erode += powf(fmax(h_ij - h_neigh,0), 2*EXP_M);
	    }

  	  amount_to_erode = -EROSION_E * powf(level_water(i,j), EXP_M) * amount_to_erode ;
	  float creep = compute_laplacian(height_landscape(i,j),
					  height_landscape(i,j-1), height_landscape(i,j+1),
					  height_landscape(i-1,j), height_landscape(i+1,j)
					  );
	  creep = EROSION_CREEP * creep;


	  float rho = 0;
	  if (level_water(i,j) > 0)
	    rho = water_times_concentration(i,j) / level_water(i,j);
	  float sedimentation = fmin(EROSION_S * rho, water_times_concentration(i,j));

	  variation_landscape(i,j) = amount_to_erode + creep + sedimentation;
	} // end for j
    } // end for i

#pragma omp barrier
  for (int i=begin_row; i<end_row; ++i)
    for (int j=0; j<NC; ++j)
      height_landscape(i,j) += DT * variation_landscape(i,j) ;
#pragma omp barrier
}



void perform_landscape_evolution(env_t &e, const unsigned int thread_id)
{
  // set border condition for water
//   set_border_condition_laplacian_zero(level_water);
//   // set border condition for landscape
//   set_border_condition_laplacian_zero(landscape);

  // set border condition for water
  // set border condition for landscape

  set_border_condition_dirichlet(e, *(e.level_water), thread_id);
//   set_border_condition_laplacian_zero(e, *(e.height_landscape), thread_id);
  set_border_condition_dirichlet(e, *(e.height_landscape), thread_id);

  // perform one step of landscape evolution
  perform_one_step_landscape_evolution(e, thread_id);
}


#endif
