#define FATAL(MSG)\
	do {\
	  fprintf(stderr, MSG "\n");\
	abort();\
	} while (0);

#include <tiffio.h>
#include <time.h>
#include <stdlib.h>
#include "io_tiff_routine.h"
#include "io_tiff_all.c"


#include <malloc.h>
#include <sstream>
#include <unistd.h>
#include <omp.h>


#include "env.hh"
#include "water_evolution.cc"
#include "landscape_evolution.cc"
// #include "sedimentation_evolution.cc"
#include "water_evolution_and_sedimentation.cc"
#include "creep.cc"
#include "save.cc"

#include "image.hh"



float integrate_landscape_over_ocean_level_thread(const env_t& e,
						  const image<float>& landscape,
						  const int thread_id)
{
  const int begin_row = e.NR_threads[thread_id];
  const int end_row   = e.NR_threads[thread_id+1];
  const int NC	      = landscape.ncols();
  float OCEAN_LEVEL = e.CST_OCEAN_LEVEL;

  float output =0;
  for (int i=begin_row; i<end_row; ++i)
    for (int j=0; j<NC; ++j)
      output+= fmax(landscape(i,j)-OCEAN_LEVEL,0.);
  return output;
}



float integrate_landscape_over_ocean_level(const env_t& e)
  {
  float output=0;
#pragma omp parallel for
  for (int thread_id=0; thread_id < e.NB_THREADS; ++thread_id)
    {
      float tmp = integrate_landscape_over_ocean_level_thread(e, *(e.height_landscape),thread_id);
      #pragma omp atomic
      output += tmp;
    }
  return output;
  }



float integrate(const env_t& e, const image<float>& ima, const int thread_id)
{
  const int begin_row = e.NR_threads[thread_id];
  const int end_row   = e.NR_threads[thread_id+1];
  const int NC	      = ima.ncols();

  float output =0;
  for (int i=begin_row; i<end_row; ++i)
    for (int j=0; j<NC; ++j)
      output+= ima(i,j);
  return output;
}



float integrate_variation_landscape(const env_t& e)
  {
  float variation=0;
#pragma omp parallel for
  for (int thread_id=0; thread_id < e.NB_THREADS; ++thread_id)
    {
      float tmp =integrate(e, *(e.variation_landscape),thread_id);
      #pragma omp atomic
      variation += tmp;
    }
  return variation;
  }


float compute_mean_level_water(const env_t& e)
  {
  float mean=0;
#pragma omp parallel for
  for (int thread_id=0; thread_id < e.NB_THREADS; ++thread_id)
    {
      float tmp =integrate(e, *(e.level_water),thread_id);
      #pragma omp atomic
      mean += tmp;
    }
  return mean/( (*(e.level_water)).nrows() * (*(e.level_water)).ncols());
  }

float compute_mean_level_landscape(const env_t& e)
  {
  float mean=0;
#pragma omp parallel for
  for (int thread_id=0; thread_id < e.NB_THREADS; ++thread_id)
    {
      float tmp =integrate(e, *(e.height_landscape),thread_id);
      #pragma omp atomic
      mean += tmp;
    }
  return mean/( (*(e.level_water)).nrows() * (*(e.level_water)).ncols());
  }



float max_over_domain(const env_t& e, const image<float>& ima, const int thread_id)
{
  const int begin_row = e.NR_threads[thread_id];
  const int end_row   = e.NR_threads[thread_id+1];
  const int NC	      = ima.ncols();

  float output =ima(begin_row,0);
  for (int i=begin_row; i<end_row; ++i)
    for (int j=0; j<NC; ++j)
      output = fmax(output, ima(i,j));
  return output;
}


float compute_max_level_water(const env_t& e)
  {
  float max=-100000;
#pragma omp parallel for
  for (int thread_id=0; thread_id < e.NB_THREADS; ++thread_id)
    {
      float tmp =max_over_domain(e, *(e.level_water),thread_id);
#pragma omp critical
      max = fmax(max, tmp);
    }
  return max;
  }


float min_over_domain(const env_t& e, const image<float>& ima, const int thread_id)
{
  const int begin_row = e.NR_threads[thread_id];
  const int end_row   = e.NR_threads[thread_id+1];
  const int NC	      = ima.ncols();

  float output =ima(begin_row,0);
  for (int i=begin_row; i<end_row; ++i)
    for (int j=0; j<NC; ++j)
      output = fmin(output, ima(i,j));
  return output;
}


float compute_min_level_water(const env_t& e)
  {
  float min=10000000000;
#pragma omp parallel for
  for (int thread_id=0; thread_id < e.NB_THREADS; ++thread_id)
    {
      float tmp =min_over_domain(e, *(e.level_water),thread_id);
#pragma omp critical
      min = fmin(min, tmp);
    }
  return min;
  }



float compute_min_level_landscape(const env_t& e)
  {
  float min=10000000000;
#pragma omp parallel for
  for (int thread_id=0; thread_id < e.NB_THREADS; ++thread_id)
    {
      float tmp =min_over_domain(e, *(e.height_landscape),thread_id);
#pragma omp critical
      min = fmin(min, tmp);
    }
  return min;
  }


float compute_max_level_landscape(const env_t& e)
  {
  float ret=-100000000;
#pragma omp parallel for
  for (int thread_id=0; thread_id < e.NB_THREADS; ++thread_id)
    {
      float tmp =max_over_domain(e, *(e.height_landscape),thread_id);
#pragma omp critical
      ret = fmax(ret, tmp);
    }
  return ret;
  }





float min_over_domain_above_ocean_level(const env_t& e, const image<float>& ima, const int thread_id)
{
  const int begin_row = e.NR_threads[thread_id];
  const int end_row   = e.NR_threads[thread_id+1];
  const int NC	      = ima.ncols();

  float output = fmax(0,ima(begin_row,0) - e.CST_OCEAN_LEVEL);
  for (int i=begin_row; i<end_row; ++i)
    for (int j=0; j<NC; ++j)
      {
	float tmp = fmax(0, ima(i,j) - e.CST_OCEAN_LEVEL);
	output = fmin(output, tmp);
      }
  return output;
}


float compute_min_level_water_above_ocean(const env_t& e)
  {
  float min=10000000000;
#pragma omp parallel for
  for (int thread_id=0; thread_id < e.NB_THREADS; ++thread_id)
    {
      float tmp =min_over_domain_above_ocean_level(e, *(e.level_water),thread_id);
#pragma omp critical
      min = fmin(min, tmp);
    }
  return min;
  }

float compute_min_level_landscape_above_ocean(const env_t& e)
  {
  float min=10000000000;
#pragma omp parallel for
  for (int thread_id=0; thread_id < e.NB_THREADS; ++thread_id)
    {
      float tmp =min_over_domain_above_ocean_level(e, *(e.height_landscape),thread_id);
#pragma omp critical
      min = fmin(min, tmp);
    }
  return min;
  }



float max_over_domain_above_ocean_level(const env_t& e, const image<float>& ima, const int thread_id)
{
  const int begin_row = e.NR_threads[thread_id];
  const int end_row   = e.NR_threads[thread_id+1];
  const int NC	      = ima.ncols();

  float output = fmax(0,ima(begin_row,0) - e.CST_OCEAN_LEVEL);
  for (int i=begin_row; i<end_row; ++i)
    for (int j=0; j<NC; ++j)
      {
	float tmp = fmax(0, ima(i,j) - e.CST_OCEAN_LEVEL);
	output = fmax(output, tmp);
      }
  return output;
}


float compute_max_level_water_above_ocean(const env_t& e)
  {
  float ret=-10000000000;
#pragma omp parallel for
  for (int thread_id=0; thread_id < e.NB_THREADS; ++thread_id)
    {
      float tmp =max_over_domain_above_ocean_level(e, *(e.level_water),thread_id);
#pragma omp critical
      ret = fmax(ret, tmp);
    }
  return ret;
  }

float compute_max_level_landscape_above_ocean(const env_t& e)
  {
  float ret=-10000000000;
#pragma omp parallel for
  for (int thread_id=0; thread_id < e.NB_THREADS; ++thread_id)
    {
      float tmp =max_over_domain_above_ocean_level(e, *(e.height_landscape),thread_id);
#pragma omp critical
      ret = fmax(ret, tmp);
    }
  return ret;
  }


float mean_over_domain_above_ocean_level(const env_t& e, const image<float>& ima, const int thread_id)
{
  const int begin_row = e.NR_threads[thread_id];
  const int end_row   = e.NR_threads[thread_id+1];
  const int NC	      = ima.ncols();

  float output = 0;
  for (int i=begin_row; i<end_row; ++i)
    for (int j=0; j<NC; ++j)
      {
	float tmp = fmax(0, ima(i,j) - e.CST_OCEAN_LEVEL);
	output += tmp;
      }
  return output;
}


float compute_mean_level_water_above_ocean(const env_t& e)
  {
  float mean=0;
#pragma omp parallel for
  for (int thread_id=0; thread_id < e.NB_THREADS; ++thread_id)
    {
      float tmp =mean_over_domain_above_ocean_level(e, *(e.level_water),thread_id);
#pragma omp critical
      mean += tmp;
    }
  return mean/( (*(e.level_water)).nrows()*(*(e.level_water)).ncols() );
  }


float compute_mean_level_landscape_above_ocean(const env_t& e)
  {
  float mean=0;
#pragma omp parallel for
  for (int thread_id=0; thread_id < e.NB_THREADS; ++thread_id)
    {
      float tmp =mean_over_domain_above_ocean_level(e, *(e.height_landscape),thread_id);
#pragma omp critical
      mean += tmp;
    }
  return mean/( (*(e.level_water)).nrows()*(*(e.level_water)).ncols() );
  }












int main2(int nrow, int ncol, float *data_in,
	  float DT_, float RAIN_,
	  float EROSION_EXP_M_, float EROSION_E_,
	  float EROSION_CREEP_, float EROSION_S_,
	  int NB_ITER_TO_PERFORM_, int NB_ITER_WATER_ONLY_,
	  float CST_OCEAN_LEVEL_,
	  float MAX_VARIATION_WATER_,
	  float PERCENTAGE_LANDSCAPE_,
	  char *OUTPUT_PATH)
{
  env_t e;



  //    image<unsigned char> orig_landscape("DEM/saltlake-8.pgm");


  const int NR= nrow;
  const int NC = ncol;
  //   image<float> orig_landscape(nrow,ncol);
  //   for (int i=0;i<NR;++i)
  //     for (int j=0;j<NC;++j)
  //       {

  //       }

  //    image<unsigned char> orig_landscape(2*64,2*64);
  //   const int NR= orig_landscape.nrows();
  //   const int NC = orig_landscape.ncols();
  //   for (int j=0;j<NC;++j)
  //     orig_landscape(NR-1,j) = 0;

  const unsigned int  NB_THREADS = omp_get_max_threads();

  const float DT = DT_;// 0.0001; alex
  const float RAIN = RAIN_;//1.875; alex
  const float EROSION_EXP_M = EROSION_EXP_M_;
  const float EROSION_E     = EROSION_E_;
  const float EROSION_CREEP = EROSION_CREEP_;
  const float EROSION_S     = EROSION_S_;
  const int NB_ITER_TO_PERFORM = NB_ITER_TO_PERFORM_;
  const int NB_ITER_WATER_ONLY = NB_ITER_WATER_ONLY_;
  const float CST_OCEAN_LEVEL = CST_OCEAN_LEVEL_; // OCEAN LEVEL
  const float MAX_VARIATION_WATER = MAX_VARIATION_WATER_;
  const float PERCENTAGE_LANDSCAPE = PERCENTAGE_LANDSCAPE_;


  std::cout << "------------------------------------------------" << std::endl;
  std::cerr << "DT= " << DT << std::endl;
  std::cerr << "rain= " << RAIN << std::endl;;
  std::cerr << "exp_M= " << EROSION_EXP_M << std::endl;;
  std::cerr << "EROSION_E= " << EROSION_E << std::endl;;
  std::cerr << "EROSION_CREEP= " << EROSION_CREEP << std::endl;;
  std::cerr << "EROSION_S= " << EROSION_S << std::endl;;
  std::cerr << "CST_OCEAN_LEVEL = " << CST_OCEAN_LEVEL << std::endl;;
  std::cerr << "iter water only = " << NB_ITER_WATER_ONLY << std::endl;;
  std::cerr << "iter to perform = " << NB_ITER_TO_PERFORM << std::endl;;
  std::cerr << "max variation water = " << MAX_VARIATION_WATER << std::endl;;
  std::cerr << "PERCENTAGE LANDSCAPE = " << PERCENTAGE_LANDSCAPE << std::endl;;


  //   const float DT = 0.01;// 0.0001; alex
  //   const unsigned int  NB_THREADS = 4;
  //   const float RAIN = DT/20.;//1.875; alex
  //   const float EROSION_EXP_M = 2;
  //   const float EROSION_E     = 1;
  //   const float EROSION_CREEP = 0;
  //   const float EROSION_S     = 0;


  omp_set_num_threads(NB_THREADS);
  std::cout << "nb_threads " << NB_THREADS << std::endl;


  create_env(e, DT, NR, NC, NB_THREADS,
	     RAIN,
	     EROSION_EXP_M, EROSION_E, EROSION_CREEP, EROSION_S,
	     CST_OCEAN_LEVEL);

  image<float> orig_landscape(NR,NC);


  for (int i=0; i<e.NR; ++i)
    for (int j=0; j<e.NC; ++j)
      {
	// 	(*(e.height_landscape))(i,j) = orig_landscape(i,j) / 256.;
	// 	(*(e.height_landscape))(i,j) = data_in[i*NC+j]/10;
	(*(e.height_landscape))(i,j) = data_in[i*NC+j]/255.;
	orig_landscape(i,j) = data_in[i*NC+j]/255.;
	// 	std::cout << data_in[i*NC+j] << " ";
	// 	if ((i>50) &&(i<e.NR-50))
	// 	  if ((j>50) && (j<e.NR-50))
	// 	    (*(e.height_landscape))(i,j) = 10;

	// set water over the landscape
	if ((*(e.height_landscape))(i,j) > e.CST_OCEAN_LEVEL)
	  (*(e.level_water))(i,j) = e.CST_OCEAN_LEVEL - ((*(e.height_landscape))(i,j));
	else
	  (*(e.level_water))(i,j) = 0.0;



	// 	(*(e.sediment))(i,j) = 0;
	(*(e.water_times_concentration))(i,j) = 0;
	(*(e.variation_landscape))(i,j) = 0;
      }

  save_image_tiff_quad_saturated(*(e.height_landscape),
				 (std::string(OUTPUT_PATH) + std::string("_orig.tif")).c_str(),
				 255);

  //perform water evolution only
  int nb_iter = 0;
  float old_mean_level = 10000000;
  float cur_mean_level = 0;
  while ((nb_iter < NB_ITER_WATER_ONLY) &&
	 (fabs(cur_mean_level-old_mean_level) > MAX_VARIATION_WATER) )
    {
#pragma omp parallel for
      for (int thread_id=0; thread_id < e.NB_THREADS; ++thread_id)
	{
	  perform_water_evolution(e, thread_id);
	}
      nb_iter++;

      if (nb_iter > 10000)
	if (nb_iter %10000 == 0)
	  {
	    old_mean_level = cur_mean_level;
	    cur_mean_level = compute_mean_level_water(e);
	  }

    } // end while

  {     // test if variation was small enough
    FILE *fp = fopen("is_variation_water_only_small_enough.txt", "w");

    if (fabs(cur_mean_level-old_mean_level) <= MAX_VARIATION_WATER)
      fprintf(fp,"OK\n");
    else
      fprintf(fp,"NOT OK\n");
    fclose(fp);
  }
  {// write number of iteration that it took
    FILE *fp = fopen("nb_iter_water_only.txt", "w");
    fprintf(fp,"%d\n",nb_iter);
    fclose(fp);
  }

  save_image_tiff_quad_saturated_visa_water(*(e.level_water),
					    (std::string(OUTPUT_PATH) +
					     std::string("_w_only.tif")).c_str(),
					    e.CST_OCEAN_LEVEL);


  {
    //compute average level of water
    float mean_level_water = compute_mean_level_water(e);
    std::cout << "mean_level_water = " << mean_level_water << std::endl;
    FILE *fp = fopen("mean_level_water_only.txt", "w");
    fprintf(fp,"%f\n",mean_level_water);
    fclose(fp);
  }

  {
    //compute max level of water
    float max_level_water = compute_max_level_water(e);
    std::cout << "max_level_water = " << max_level_water << std::endl;
    FILE *fp = fopen("max_level_water_only.txt", "w");
    fprintf(fp,"%f\n",max_level_water);
    fclose(fp);
  }

  {
    //compute min level of water
    float min_level_water = compute_min_level_water(e);
    std::cout << "min_level_water = " << min_level_water << std::endl;
    FILE *fp = fopen("min_level_water_only.txt", "w");
    fprintf(fp,"%f\n",min_level_water);
    fclose(fp);
  }



  {
    //compute average level of water above ocean
    float mean_level_water = compute_mean_level_water_above_ocean(e);
    std::cout << "mean_level_water = " << mean_level_water << std::endl;
    FILE *fp = fopen("mean_level_water_above_ocean_water_only.txt", "w");
    fprintf(fp,"%f\n",mean_level_water);
    fclose(fp);
  }

  {
    //compute max level of water  above ocean
    float max_level_water = compute_max_level_water_above_ocean(e);
    std::cout << "max_level_water = " << max_level_water << std::endl;
    FILE *fp = fopen("max_level_water_above_ocean_water_only.txt", "w");
    fprintf(fp,"%f\n",max_level_water);
    fclose(fp);
  }

  {
    //compute min level of water  above ocean
    float min_level_water = compute_min_level_water_above_ocean(e);
    std::cout << "min_level_water = " << min_level_water << std::endl;
    FILE *fp = fopen("min_level_water_above_ocean_water_only.txt", "w");
    fprintf(fp,"%f\n",min_level_water);
    fclose(fp);
  }






  // ***************************************************************************
  // ***************************************************************************
  // *************** Water evolution only is finished **************************
  // ***************************************************************************
  // ***************************************************************************

  nb_iter = 0;

  FILE *fp = fopen("percentage_landscape_removed_iteration.txt", "w");

  float target_mass_removed = integrate_landscape_over_ocean_level(e);
  target_mass_removed *= (1.-PERCENTAGE_LANDSCAPE);
  float current_mass_landscape = integrate_landscape_over_ocean_level(e);
  while ( (current_mass_landscape >= target_mass_removed) && (nb_iter <NB_ITER_TO_PERFORM))
    {
#pragma omp parallel for
      for (int thread_id=0; thread_id < e.NB_THREADS; ++thread_id)
	perform_landscape_evolution(e, thread_id);

      for (int thread_id=0; thread_id < e.NB_THREADS; ++thread_id)
	perform_creep_evolution(e, thread_id);


#pragma omp parallel for
      for (int thread_id=0; thread_id < e.NB_THREADS; ++thread_id)
	perform_water_and_sedimentation_evolution(e, thread_id);
      if (nb_iter%5000 == 0)
	{
	  current_mass_landscape = integrate_landscape_over_ocean_level(e);
	  std::cout << current_mass_landscape << " " << target_mass_removed << std::endl;
	  fprintf(fp, "iter=%d current mass=%f  target mass= %f\n",
		  nb_iter, current_mass_landscape, target_mass_removed);
	}
      nb_iter++;
    } //end of while
  fclose(fp);
  // END ITERATION FULL EQUATION


  image<float> diff_landscape((*(e.level_water)).nrows(),(*(e.level_water)).ncols());
  for (int i=0; i<e.NR; ++i)
    for (int j=0; j<e.NC; ++j)
      diff_landscape(i,j) = orig_landscape(i,j) - (*(e.height_landscape))(i,j);


  image<float> concentration(e.NR, e.NC);
  for (int i=0; i<e.NR; ++i)
    for (int j=0; j<e.NC; ++j)
      concentration(i,j) = (*(e.water_times_concentration))(i,j)/(0.0000001+(*(e.level_water))(i,j));




  {     // test if variation was small enough
    FILE *fp = fopen("is_landscape_removed_ok.txt", "w");

    if(current_mass_landscape < target_mass_removed)
      fprintf(fp,"OK\n");
    else
      fprintf(fp,"NOT OK\n");
    fclose(fp);
  }
  {// write number of iteration that it took
    FILE *fp = fopen("nb_iter_full_equations.txt", "w");
    fprintf(fp,"%d\n",nb_iter);
    fclose(fp);
  }


  // compute some statistique
  {
    {
      //compute average level of water
      float mean_level_water = compute_mean_level_water(e);
      std::cout << "mean_level_water = " << mean_level_water << std::endl;
      FILE *fp = fopen("mean_level_water_full.txt", "w");
      fprintf(fp,"%f\n",mean_level_water);
      fclose(fp);
    }

    {
      //compute max level of water
      float max_level_water = compute_max_level_water(e);
      std::cout << "max_level_water = " << max_level_water << std::endl;
      FILE *fp = fopen("max_level_water_full.txt", "w");
      fprintf(fp,"%f\n",max_level_water);
      fclose(fp);
    }

    {
      //compute min level of water
      float min_level_water = compute_min_level_water(e);
      std::cout << "min_level_water = " << min_level_water << std::endl;
      FILE *fp = fopen("min_level_water_full.txt", "w");
      fprintf(fp,"%f\n",min_level_water);
      fclose(fp);
    }

    {
      //compute average level of water above ocean
      float mean_level_water = compute_mean_level_water_above_ocean(e);
      std::cout << "mean_level_water = " << mean_level_water << std::endl;
      FILE *fp = fopen("mean_level_water_above_ocean_full.txt", "w");
      fprintf(fp,"%f\n",mean_level_water);
      fclose(fp);
    }

    {
      //compute max level of water  above ocean
      float max_level_water = compute_max_level_water_above_ocean(e);
      std::cout << "max_level_water = " << max_level_water << std::endl;
      FILE *fp = fopen("max_level_water_above_ocean_full.txt", "w");
      fprintf(fp,"%f\n",max_level_water);
      fclose(fp);
    }

    {
      //compute min level of water  above ocean
      float min_level_water = compute_min_level_water_above_ocean(e);
      std::cout << "min_level_water = " << min_level_water << std::endl;
      FILE *fp = fopen("min_level_water_above_ocean_full.txt", "w");
      fprintf(fp,"%f\n",min_level_water);
      fclose(fp);
    }



    // end stat water

    {
      //compute average level of landscape
      float mean_level_landscape = compute_mean_level_landscape(e);
      std::cout << "mean_level_landscape = " << mean_level_landscape << std::endl;
      FILE *fp = fopen("mean_level_landscape_full.txt", "w");
      fprintf(fp,"%f\n",mean_level_landscape);
      fclose(fp);
    }

    {
      //compute max level of landscape
      float max_level_landscape = compute_max_level_landscape(e);
      std::cout << "max_level_landscape = " << max_level_landscape << std::endl;
      FILE *fp = fopen("max_level_landscape_full.txt", "w");
      fprintf(fp,"%f\n",max_level_landscape);
      fclose(fp);
    }

    {
      //compute min level of landscape
      float min_level_landscape = compute_min_level_landscape(e);
      std::cout << "min_level_landscape = " << min_level_landscape << std::endl;
      FILE *fp = fopen("min_level_landscape_full.txt", "w");
      fprintf(fp,"%f\n",min_level_landscape);
      fclose(fp);
    }


    {
      //compute average level of landscape above ocean
      float mean_level_landscape = compute_mean_level_landscape_above_ocean(e);
      std::cout << "mean_level_landscape = " << mean_level_landscape << std::endl;
      FILE *fp = fopen("mean_level_landscape_above_ocean_full.txt", "w");
      fprintf(fp,"%f\n",mean_level_landscape);
      fclose(fp);
    }

    {
      //compute max level of landscape above ocean
      float max_level_landscape = compute_max_level_landscape_above_ocean(e);
      std::cout << "max_level_landscape = " << max_level_landscape << std::endl;
      FILE *fp = fopen("max_level_landscape_above_ocean_full.txt", "w");
      fprintf(fp,"%f\n",max_level_landscape);
      fclose(fp);
    }

    {
      //compute min level of landscape above ocean
      float min_level_landscape = compute_min_level_landscape_above_ocean(e);
      std::cout << "min_level_landscape = " << min_level_landscape << std::endl;
      FILE *fp = fopen("min_level_landscape_above_ocean_full.txt", "w");
      fprintf(fp,"%f\n",min_level_landscape);
      fclose(fp);
    }

    // end stat landscape

    float mean_level_difference_landscape = 0;
    float max_level_difference_landscape = -10000;
    float min_level_difference_landscape = 10000;

    for (int i=0; i<e.NR; ++i)
      for (int j=0; j<e.NC; ++j)
	{
	  mean_level_difference_landscape += diff_landscape(i,j);
	  max_level_difference_landscape = fmax(max_level_difference_landscape, diff_landscape(i,j));
	  min_level_difference_landscape = fmin(min_level_difference_landscape, diff_landscape(i,j));
	}
    mean_level_difference_landscape /= (e.NR*e.NC);

    //compute variance
    float var_diff_landscape = 0;
    for (int i=0; i<e.NR; ++i)
      for (int j=0; j<e.NC; ++j)
	var_diff_landscape +=
	  (diff_landscape(i,j) - mean_level_difference_landscape)*
	  (diff_landscape(i,j) - mean_level_difference_landscape);
    var_diff_landscape = var_diff_landscape / (e.NR*e.NC-1);
    float sigma_diff_landscape = sqrt(var_diff_landscape);


    {
      FILE *fp = fopen("mean_level_difference_landscape_full.txt", "w");
      fprintf(fp,"%f\n",mean_level_difference_landscape);
      fclose(fp);
    }
    {
      FILE *fp = fopen("sigma_level_difference_landscape_full.txt", "w");
      fprintf(fp,"%f\n",sigma_diff_landscape);
      fclose(fp);
    }

    {
      FILE *fp = fopen("max_level_difference_landscape_full.txt", "w");
      fprintf(fp,"%f\n",max_level_difference_landscape);
      fclose(fp);
    }
    {
      FILE *fp = fopen("min_level_difference_landscape_full.txt", "w");
      fprintf(fp,"%f\n",min_level_difference_landscape);
      fclose(fp);
    }

    // end difference landscape

    {
      float mean_level_concentration = 0;
      float max_level_concentration = -10000;
      float min_level_concentration = 10000;

      for (int i=0; i<e.NR; ++i)
	for (int j=0; j<e.NC; ++j)
	  {
	    mean_level_concentration += concentration(i,j);
	    max_level_concentration = fmax(max_level_concentration, concentration(i,j));
	    min_level_concentration = fmin(min_level_concentration, concentration(i,j));
	  }
      mean_level_concentration /= (e.NR*e.NC);


      //compute variance
      float var_concentration = 0;
      for (int i=0; i<e.NR; ++i)
	for (int j=0; j<e.NC; ++j)
	  var_concentration +=
	    (concentration(i,j) - mean_level_concentration)*
	    (concentration(i,j) - mean_level_concentration);
      var_concentration = var_concentration / (e.NR*e.NC-1);
      float sigma_var_concentration = sqrt(var_concentration);


      {
	FILE *fp = fopen("mean_level_concentration_full.txt", "w");
	fprintf(fp,"%f\n",mean_level_concentration);
	fclose(fp);
      }
      {
	FILE *fp = fopen("sigma_level_concentration_full.txt", "w");
	fprintf(fp,"%f\n",sigma_var_concentration);
	fclose(fp);
      }
      {
	FILE *fp = fopen("max_level_concentration_full.txt", "w");
	fprintf(fp,"%f\n",max_level_concentration);
	fclose(fp);
      }
      {
	FILE *fp = fopen("min_level_concentration_full.txt", "w");
	fprintf(fp,"%f\n",min_level_concentration);
	fclose(fp);
      }

    }


  } // end of statistique


  {
    save_image_tiff_quad_saturated_visa_water(*(e.level_water),
					      (std::string(OUTPUT_PATH) +
					       std::string("_w.tif")).c_str(),
					      e.CST_OCEAN_LEVEL);


    save_image_tiff_quad_saturated(*(e.height_landscape),
				   (std::string(OUTPUT_PATH) + std::string("_l.tif")).c_str(), 255);

//     save_image_tiff_quad_saturated_old_minus_new_landscape(orig_landscape,
// 							   *(e.height_landscape),
// 							   (std::string(OUTPUT_PATH)
// 							    + std::string("_diff_landscape.tif")).c_str(), 255);

    save_image_tiff_quad_saturated_old_minus_new_landscape(diff_landscape,
							   (std::string(OUTPUT_PATH)
							    + std::string("_diff_landscape.tif")).c_str());


    save_image_tiff_quad_saturated_concentration(concentration,
						 (std::string(OUTPUT_PATH) + std::string("_c.tif")).c_str()
						 );


    //      save_image_tiff(*(e.level_water),
    // 		     (std::string("w") + std::string(OUTPUT_PATH)).c_str(), 256);
  }





  destroy_env(e);
  return 0;
}








// int main2(int ac, char* av[])
// {
//   std::cerr << av[2] << std::endl;
//   std::cerr << av[3] << std::endl;

//   image<unsigned char> in(av[2]);
//   in.save(av[3]);
//   return 0;
// }



int main(int ac, char* av[])
{


  float        *data_in;                    /* input data */
  float            *ptr_in_rgb[3];
  size_t       n_col, n_row;                /* number of columns and rows */
  size_t       n_channels;                  /* number of channels */
  unsigned int k;

  /*
    * argv[1] contains the input file name
    * argv[2] contains the output file name
    * argv[3] time_step
    * argv[4] rain
    * argv[5] exp_M
    * argv[6] erosion_e
    * argv[7] erosion_creep
    * argv[8] erosion_s
    * argv[9] iter_for_save
    */
//   std::cout << "------------------------------------------------" << std::endl;
//   std::cerr << "DT= " << atof(av[3]) << std::endl;
//   std::cerr << "rain= " << atof(av[4]) << std::endl;;
//   std::cerr << "exp_M= " << atof(av[5]) << std::endl;;
//   std::cerr << "EROSION_E= " << atof(av[6]) << std::endl;;
//   std::cerr << "EROSION_CREEP= " << atof(av[7]) << std::endl;;
//   std::cerr << "EROSION_S= " << atof(av[8]) << std::endl;;
//   std::cerr << "iter_for_save= " << atoi(av[9]) << std::endl;;

  if (ac != 14)
    return 1;

  const float DT = atof(av[3]);
  const float RAIN = atof(av[4]) * 0.0000001;
  const float EROSION_EXP_M = atof(av[5]);
  const float EROSION_E     = atof(av[6]);
  const float EROSION_CREEP = atof(av[7]);
  const float EROSION_S     = atof(av[8]);
  const int   NB_ITER = atoi(av[9]) * 1000;
  const int   NB_ITER_WATER_ONLY = atoi(av[10]) * 100;
  const float CST_OCEAN_LEVEL = atof(av[11]);
  const float MAX_VARIATION_WATER = atof(av[12]);
  const float PERCENTAGE_LANDSCAPE = atof(av[13])/100.; // to get a percentage


   /* read the TIFF image */
   if (NULL == (data_in = read_tiff_f32(av[1], &n_col, &n_row, &n_channels)))
      FATAL("error while reading the TIFF image");



   ptr_in_rgb[0] = data_in;
   ptr_in_rgb[1] = data_in + n_row * n_col;
   ptr_in_rgb[2] = data_in + 2 * n_row * n_col;

   // check gray
   for (k=0; k<(n_row*n_col-1); k++)
     if((*(ptr_in_rgb[0]+k)!= *(ptr_in_rgb[1]+k))||(*(ptr_in_rgb[0]+k)!=*(ptr_in_rgb[2]+k))){
       n_channels=3;
       std::cout << " ERROR NOT GRAY " << std::endl;
     }



//    if (n_channels != 1)
//      FATAL("allocation error");





//    write_tiff_f32((std::string(av[2]) + std::string("_l.tif")).c_str(), data_in, n_col, n_row, n_channels);
//    write_tiff_f32((std::string(av[2]) + std::string("_w.tif")).c_str(), data_in, n_col, n_row, n_channels);
//    return 0;

   main2(n_row, n_col, data_in,
	 DT, RAIN, EROSION_EXP_M, EROSION_E, EROSION_CREEP, EROSION_S,
	 NB_ITER, NB_ITER_WATER_ONLY,
	 CST_OCEAN_LEVEL, MAX_VARIATION_WATER,
	 PERCENTAGE_LANDSCAPE,
	 av[2]);
      return 0;
}
