#ifndef __SEDIMENT_EVOLUTION_HH__
#define __SEDIMENT_EVOLUTION_HH__

#include <cmath>

#include "env.hh"
#include "image.hh"
#include "border_condition.cc"


// // current site s
// // neighboring site t
inline
float get_variation_sediment_level_for_current_site(const float landscape_s,
						    const float water_s,
						    const float water_times_concentration_s,
						    const float landscape_t,
						    const float water_t,
						    const float water_times_concentration_t
						    )
{
  float steep = (landscape_t + water_t) -  (landscape_s + water_s);
  if (steep > 0) // t is higher
    return water_times_concentration_t * steep; // so s receives water
  return water_times_concentration_s * steep; // so s sends water
}


// return time step used during the iiteration
void perform_one_step_sediment_evolution(env_t &e, const unsigned int thread_id)
{
  image<float>& level_water			= *(e.level_water);
  image<float>& height_landscape		= *(e.height_landscape);
  image<float>& water_times_concentration	= *(e.water_times_concentration);
  image<float>& variation_landscape		= *(e.variation_landscape);
  image<float>& delta_water			= *(e.delta_water);

  const float DT				= e.DT;

  const int NB_W = 4;
  const int win_i[NB_W] = {0,  0, -1, 1};
  const int win_j[NB_W] = {-1, 1,  0, 0};

  const int begin_row = e.NR_threads[thread_id];
  const int end_row   = e.NR_threads[thread_id+1];
  const int NC		= level_water.ncols();

  for (int i=begin_row; i<end_row; ++i)
    for (int j=0; j<NC; ++j)
      delta_water(i,j) = 0;

  // compute variations
  for (int i=begin_row; i<end_row; ++i)
    for (int j=0; j<NC; ++j)
      {
	float delta_water_cur = 0;
	for (int k=0; k<NB_W; ++k)
	  {
	    int n_i = i+win_i[k];
	    int n_j = j+win_j[k];
	    delta_water_cur +=
	      get_variation_sediment_level_for_current_site(height_landscape(i,j),
							    level_water(i,j),
							    water_times_concentration(i,j),
							    height_landscape(n_i,n_j),
							    level_water(n_i,n_j),
							    water_times_concentration(n_i,n_j)
							    );
	  } // end iterate neighbor
	delta_water(i,j) = delta_water_cur;
      }


#pragma omp barrier
  for (int i=begin_row; i<end_row; ++i)
    for (int j=0; j<NC; ++j)
      {
	water_times_concentration(i,j) += DT*(delta_water(i,j) - variation_landscape(i,j));
// 	water_times_concentration(i,j) = fmax(water_times_concentration(i,j), 0);
      }
#pragma omp barrier
}



void perform_sediment_evolution(env_t &e, const unsigned int thread_id)
{
  // set border condition for water
  set_border_condition_dirichlet(e, *(e.level_water), thread_id);
  // set border condition for landscape
//   set_border_condition_laplacian_zero(e, *(e.height_landscape), thread_id);
  set_border_condition_dirichlet(e, *(e.height_landscape), thread_id);

  set_border_condition_dirichlet(e, *(e.water_times_concentration), thread_id);

  // perform one step of water evolution
  perform_one_step_sediment_evolution(e, thread_id);
}

#endif
