#ifndef __WATER_AND_SEDIMENTATION_EVOLUTION_HH__
#define __WATER_AND_SEDIMENTATION_EVOLUTION_HH__

#include <cmath>

#include "image.hh"
#include "border_condition.cc"


// current site s
// neighboring site t
inline
void get_variation_water_level_and_steep_for_current_site(const float landscape_s,
							  const float water_s,
							  const float water_times_concentration_s,
							  const float landscape_t,
							  const float water_t,
							  const float water_times_concentration_t,
							  float & val_for_water,
							  float & val_for_concentration)
{
  float steep = (landscape_t + water_t) -  (landscape_s + water_s);
  if (steep > 0) // t is higher
    {
      val_for_water         =  water_t * steep; // so s receives water
      val_for_concentration =  water_times_concentration_t * steep;
    }
  else
    {
      val_for_water         = water_s * steep; // so s sends water
      val_for_concentration = water_times_concentration_s * steep;
    }
  return ;
}


// return time step used during the iiteration
void perform_one_step_water_and_sedimentation_evolution(env_t &e, const unsigned int thread_id)
{
  image<float>& level_water       = *(e.level_water);
  image<float>& height_landscape  = *(e.height_landscape);
  image<float>& delta_water	  = *(e.delta_water);
  image<float>& water_times_concentration	  = *(e.water_times_concentration);
  image<float>& delta_water_times_concentration	  = *(e.delta_water_times_concentration);

  // variation landscape = (z^(n+1)-z^(n)) without creep
  image<float>& variation_landscape = *(e.variation_landscape);
  const float RAIN                = e.RAIN;
  const float DT                  = e.DT;


  const int begin_row = e.NR_threads[thread_id];
  const int end_row   = e.NR_threads[thread_id+1];
  const int NC		= level_water.ncols();

  // compute variations
  for (int i=begin_row; i<end_row; ++i)
    for (int j=0; j<NC; ++j)
      {
	float delta_water_cur = 0;
	float delta_concentration_time_water_cur = 0;
	float val_for_water = 0;
	float val_for_concentration = 0;

	get_variation_water_level_and_steep_for_current_site(height_landscape(i,j),
							     level_water     (i,j),
							     water_times_concentration(i,j),
							     height_landscape(i,j-1),
							     level_water     (i,j-1),
							     water_times_concentration(i,j-1),
							     val_for_water,
							     val_for_concentration);
	delta_water_cur += val_for_water;
	delta_concentration_time_water_cur += val_for_concentration;

	get_variation_water_level_and_steep_for_current_site(height_landscape(i,j),
							     level_water     (i,j),
							     water_times_concentration(i,j),
							     height_landscape(i,j+1),
							     level_water     (i,j+1),
							     water_times_concentration(i,j+1),
							     val_for_water,
							     val_for_concentration);
	delta_water_cur += val_for_water;
	delta_concentration_time_water_cur += val_for_concentration;
	get_variation_water_level_and_steep_for_current_site(height_landscape(i,j),
							     level_water     (i,j),
							     water_times_concentration(i,j),
							     height_landscape(i-1,j),
							     level_water     (i-1,j),
							     water_times_concentration(i-1,j),
							     val_for_water,
							     val_for_concentration);
	delta_water_cur += val_for_water;
	delta_concentration_time_water_cur += val_for_concentration;
	get_variation_water_level_and_steep_for_current_site(height_landscape(i,j),
							     level_water     (i,j),
							     water_times_concentration(i,j),
							     height_landscape(i+1,j),
							     level_water     (i+1,j),
							     water_times_concentration(i+1,j),
							     val_for_water,
							     val_for_concentration);
	delta_water_cur += val_for_water;
	delta_concentration_time_water_cur += val_for_concentration;

	// update global value
	delta_water(i,j) = delta_water_cur;
	delta_water_times_concentration(i,j) = delta_concentration_time_water_cur;

      }


#pragma omp barrier
  for (int i=begin_row; i<end_row; ++i)
    for (int j=0; j<NC; ++j)
      {
	level_water(i,j) += DT*(delta_water(i,j) + RAIN);
	// recall variation_landscape is z^(n+1) - z(n) without the creep
	water_times_concentration(i,j) +=
	  DT*(delta_water_times_concentration(i,j)) - variation_landscape(i,j);
      }
#pragma omp barrier
}



void perform_water_and_sedimentation_evolution(env_t &e, const unsigned int thread_id)
{
  // set border condition for water
//   set_border_condition_dirichlet(e, *(e.level_water), thread_id);
  // set border condition for landscape
//   set_border_condition_laplacian_zero(e, *(e.height_landscape), thread_id);
//   set_border_condition_dirichlet(e, *(e.height_landscape), thread_id);


//    set_border_condition_laplacian_zero(e, *(e.level_water), thread_id);
//    set_border_condition_dirichlet(e, *(e.level_water), thread_id);

  set_border_condition_ocean(e,
			     *(e.level_water),
			     *(e.height_landscape),
			     e.CST_OCEAN_LEVEL,
			     thread_id);

  // perform one step of water evolution
  perform_one_step_water_and_sedimentation_evolution(e, thread_id);
}

#endif
